# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# complementary header for C++ STL libs not included in Cython

from libc.stdint cimport uint_fast64_t


cdef extern from "<random>" namespace "std" nogil:
    cdef cppclass mt19937_64:
        ctypedef uint_fast64_t result_type

        mt19937_64() except +
        mt19937_64(result_type seed) except +
        result_type operator()() except +
        result_type min() except +
        result_type max() except +
        void discard(size_t z) except +
        void seed(result_type seed) except +



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import concurrent.futures as futures
import functools
import logging
import os
import weakref
from typing import Any, Callable, Coroutine, Dict, List, Optional, Tuple, Type

import cloudpickle
import numpy as np

from ...nvutils import get_cuda_context, get_index_and_uuid
from ...serialization import deserialize
from ...serialization.aio import BUFFER_SIZES_NAME, AioSerializer, get_header_length
from ...utils import classproperty, implements, is_cuda_buffer, is_v6_ip, lazy_import
from ..message import _MessageBase
from .base import Channel, ChannelType, Client, Server
from .core import register_client, register_server
from .errors import ChannelClosed

ucp = lazy_import("ucxx")
numba_cuda = lazy_import("numba.cuda")
rmm = lazy_import("rmm")

_warning_suffix = (
    "This is often the result of a CUDA-enabled library calling a CUDA runtime function before "
    "spawning worker processes. Please make sure any such function calls don't happen "
    "at import time or in the global scope of a program."
)


logger = logging.getLogger(__name__)


def synchronize_stream(stream: int = 0):
    ctx = numba_cuda.current_context()
    cu_stream = numba_cuda.driver.drvapi.cu_stream(stream)
    stream = numba_cuda.driver.Stream(ctx, cu_stream, None)
    stream.synchronize()  # type: ignore


class UCXInitializer:
    _inited = False

    @staticmethod
    def _get_options(ucx_config: dict) -> Tuple[dict, dict]:
        """
        Get options and envs from ucx options in oscar config
        """
        options = dict()
        envs = dict()

        # if any of the flags are set, as long as they are not Null/None,
        # we assume we should configure basic TLS settings for UCX, otherwise we
        # leave UCX to its default configuration
        if any(ucx_config.get(name) for name in ["tcp", "nvlink", "infiniband"]):
            if ucx_config.get("rdmacm"):  # pragma: no cover
                tls = "tcp"
                tls_priority = "rdmacm"
            else:
                tls = "tcp"
                tls_priority = "tcp"

            # CUDA COPY can optionally be used with ucx -- we rely on the user
            # to define when messages will include CUDA objects.  Note:
            # defining only the Infiniband flag will not enable cuda_copy
            if any(
                ucx_config.get(name) for name in ["nvlink", "cuda-copy"]
            ):  # pragma: no cover
                tls += ",cuda_copy"

            if ucx_config.get("infiniband"):  # pragma: no cover
                tls = "ib," + tls
            if ucx_config.get("nvlink"):  # pragma: no cover
                tls += ",cuda_ipc"

            options["TLS"] = tls
            options["SOCKADDR_TLS_PRIORITY"] = tls_priority
        elif "UCX_TLS" in os.environ:  # pragma: no cover
            options["TLS"] = os.environ["UCX_TLS"]

        for k, v in ucx_config.get("environment", dict()).items():  # pragma: no cover
            # {"some-name": value} is translated to {"UCX_SOME_NAME": value}
            key = f'UCX_{"_".join(s.upper() for s in k.split("-"))}'
            opt_key = key[4:]
            if opt_key in options:
                logger.warning(
                    f"Ignoring {k}={v} (key={key}) in ucx.environment, "
                    f"preferring {opt_key}={options[opt_key]} "
                    "from high level options"
                )
            elif key in os.environ:
                # This is only info because setting UCX configuration via
                # environment variables is a reasonably common approach
                logger.info(
                    f"Ignoring {k}={v} (key={key}) in ucx.environment, "
                    f"preferring {key}={os.environ[key]} from external environment"
                )
            else:
                envs[key] = v

        return options, envs

    @staticmethod
    def init(ucx_config: dict):
        if UCXInitializer._inited:
            return

        options, envs = UCXInitializer._get_options(ucx_config)

        # We ensure the CUDA context is created before initializing UCX. This can't
        # be safely handled externally because communications start before
        # preload scripts run.
        # Precedence:
        # 1. external environment
        # 2. ucx_config (high level settings passed to ucp.init)
        # 3. ucx_environment (low level settings equivalent to environment variables)
        ucx_tls = os.environ.get("UCX_TLS", options.get("TLS", envs.get("UCX_TLS", "")))
        if (
            ucx_config.get("create-cuda-contex") is True
            # This is not foolproof, if UCX_TLS=all we might require CUDA
            # depending on configuration of UCX, but this is better than
            # nothing
            or ("cuda" in ucx_tls and "^cuda" not in ucx_tls)
        ):
            if numba_cuda is None:  # pragma: no cover
                raise ImportError(
                    "CUDA support with UCX requires Numba for context management"
                )

            pre_existing_cuda_context = get_cuda_context()
            if pre_existing_cuda_context.has_context:
                dev = pre_existing_cuda_context.device_info
                assert dev is not None
                logger.warning(
                    f"A CUDA context for device {dev.device_index} ({str(dev.uuid)}) "
                    f"already exists on process ID {os.getpid()}. {_warning_suffix}"
                )

            numba_cuda.current_context()

            cuda_context_created = get_cuda_context()
            cuda_visible_device = get_index_and_uuid(
                os.environ.get("CUDA_VISIBLE_DEVICES", "0").split(",")[0]
            )
            if (
                cuda_context_created.has_context
                and cuda_context_created.device_info.uuid != cuda_visible_device.uuid  # type: ignore
            ):  # pragma: no cover
                cuda_context_created_dev = cuda_context_created.device_info
                assert cuda_context_created_dev is not None
                logger.warning(
                    f"Worker with process ID {os.getpid()} should have a CUDA context assigned to device "
                    f"{cuda_visible_device.device_index} ({str(cuda_visible_device.uuid)}), "  # type: ignore
                    f"but instead the CUDA context is on device {cuda_context_created_dev.device_index} "
                    f"({str(cuda_context_created_dev.uuid)}). {_warning_suffix}"
                )

        original_environ = os.environ
        new_environ = os.environ.copy()
        new_environ.update(envs)
        os.environ = new_environ  # type: ignore
        try:
            # let UCX determine the appropriate transports
            ucp.init()
        finally:
            os.environ = original_environ

        UCXInitializer._inited = True

    @staticmethod
    def reset():
        ucp.reset()
        UCXInitializer._inited = False


class UCXChannel(Channel):
    __slots__ = (
        "ucp_endpoint",
        "_closed",
        "_has_close_callback",
        "_send_lock",
        "_recv_lock",
        "__weakref__",
    )

    name = "ucx"

    def __init__(
        self,
        ucp_endpoint: "ucp.Endpoint",  # type: ignore
        local_address: str | None = None,
        dest_address: str | None = None,
        compression: str | None = None,
    ):
        super().__init__(
            local_address=local_address,
            dest_address=dest_address,
            compression=compression,
        )
        self.ucp_endpoint = ucp_endpoint

        self._send_lock = asyncio.Lock()
        self._recv_lock = asyncio.Lock()

        # When the UCX endpoint closes or errors the registered callback
        # is called.
        if hasattr(self.ucp_endpoint, "set_close_callback"):
            ref = weakref.ref(self)
            self.ucp_endpoint.set_close_callback(
                functools.partial(UCXChannel._close_channel, ref)
            )
            self._closed = False
            self._has_close_callback = True
        else:  # pragma: no cover
            self._has_close_callback = False

    @staticmethod
    def _close_channel(channel_ref: weakref.ReferenceType):
        channel = channel_ref()
        if channel is not None:
            channel._closed = True

    async def _serialize(self, message: Any) -> List[bytes]:
        compress = self.compression or 0
        serializer = AioSerializer(message, compress=compress)
        return await serializer.run()

    @property
    @implements(Channel.type)
    def type(self) -> int:
        return ChannelType.remote

    @implements(Channel.send)
    async def send(self, message: Any):
        if self.closed:
            raise ChannelClosed("UCX Endpoint is closed, unable to send message")

        buffers = await self._serialize(message)
        return await self.send_buffers(buffers)

    @implements(Channel.recv)
    async def recv(self):
        async with self._recv_lock:
            try:
                info_buffer = np.empty(11, dtype="u1").data
                await self.ucp_endpoint.recv(info_buffer)
                head_length = get_header_length(info_buffer)
                header_buffer = np.empty(head_length, dtype="u1").data
                await self.ucp_endpoint.recv(header_buffer)
                header = cloudpickle.loads(header_buffer)

                is_cuda_buffers = header[0].get("is_cuda_buffers")
                buffer_sizes = header[0].pop(BUFFER_SIZES_NAME)

                buffers = []
                for is_cuda_buffer, buf_size in zip(is_cuda_buffers, buffer_sizes):
                    if buf_size == 0:  # pragma: no cover
                        buffers.append(bytes())
                    elif is_cuda_buffer:
                        cuda_buffer = rmm.DeviceBuffer(size=buf_size)
                        await self.ucp_endpoint.recv(cuda_buffer)
                        buffers.append(cuda_buffer)
                    else:
                        buffer = np.empty(buf_size, dtype="u1").data
                        await self.ucp_endpoint.recv(buffer)
                        buffers.append(buffer)
            except BaseException as e:
                if not self._closed:
                    # In addition to UCX exceptions, may be CancelledError or another
                    # "low-level" exception. The only safe thing to do is to abort.
                    self.abort()
                    raise ChannelClosed(
                        f"Connection closed by writer.\nInner exception: {e!r}"
                    ) from e
                else:
                    raise EOFError("Server closed already")
        return deserialize(header, buffers)

    async def send_buffers(self, buffers: list, meta: Optional[_MessageBase] = None):
        try:
            # It is necessary to first synchronize the default stream before start
            # sending We synchronize the default stream because UCX is not
            # stream-ordered and syncing the default stream will wait for other
            # non-blocking CUDA streams. Note this is only sufficient if the memory
            # being sent is not currently in use on non-blocking CUDA streams.
            if any(is_cuda_buffer(buf) for buf in buffers):
                # has GPU buffer
                synchronize_stream(0)

            meta_buffers = None
            if meta:
                meta_buffers = await self._serialize(meta)

            async with self._send_lock:
                if meta_buffers:
                    for buf in meta_buffers:
                        await self.ucp_endpoint.send(buf)
                for buffer in buffers:
                    await self.ucp_endpoint.send(buffer)
        except ucp.exceptions.UCXError:  # pragma: no cover
            self.abort()
            raise ChannelClosed("While writing, the connection was closed")

    async def recv_buffers(self, buffers: list):
        async with self._recv_lock:
            try:
                for buffer in buffers:
                    await self.ucp_endpoint.recv(buffer)
            except BaseException as e:  # pragma: no cover
                if not self._closed:
                    # In addition to UCX exceptions, may be CancelledError or another
                    # "low-level" exception. The only safe thing to do is to abort.
                    self.abort()
                    raise ChannelClosed(
                        f"Connection closed by writer.\nInner exception: {e!r}"
                    ) from e
                else:
                    raise EOFError("Server closed already")

    def abort(self):
        self._closed = True
        if self.ucp_endpoint is not None:
            self.ucp_endpoint.abort()
            self.ucp_endpoint = None

    @implements(Channel.close)
    async def close(self):
        self._closed = True
        if self.ucp_endpoint is not None:
            await self.ucp_endpoint.close()
            # abort
            self.ucp_endpoint.abort()
            self.ucp_endpoint = None

    @property
    @implements(Channel.closed)
    def closed(self):
        if self._has_close_callback is None:  # pragma: no cover
            # The self._closed flag is separate from the endpoint's lifetime, even when
            # the endpoint has closed or errored, there may be messages on its buffer
            # still to be received, even though sending is not possible anymore.
            return self._closed
        else:
            return self.ucp_endpoint is None


@register_server
class UCXServer(Server):
    __slots__ = "host", "port", "_ucp_listener", "_channels", "_closed"

    scheme = "ucx"

    _ucp_listener: "ucp.Listener"  # type: ignore
    _channels: set[UCXChannel]

    def __init__(
        self,
        host: str,
        port: int,
        ucp_listener: "ucp.Listener",  # type: ignore
        channel_handler: Callable[[Channel], Coroutine] | None = None,
    ):
        super().__init__(f"{UCXServer.scheme}://{host}:{port}", channel_handler)
        self.host = host
        self.port = port
        self._ucp_listener = ucp_listener
        self._channels = set()
        self._closed = asyncio.Event()

    @classproperty
    @implements(Server.client_type)
    def client_type(self) -> Type["Client"]:
        return UCXClient

    @property
    @implements(Server.channel_type)
    def channel_type(self) -> int:
        return ChannelType.remote

    @staticmethod
    async def create(config: Dict) -> "Server":
        config = config.copy()
        if "address" in config:
            address = config.pop("address")
            prefix = f"{UCXServer.scheme}://"
            if address.startswith(prefix):
                address = address[len(prefix) :]
            host, port = address.rsplit(":", 1)
            port = int(port)
        else:
            host = config.pop("host")
            port = int(config.pop("port"))
        _host = host
        if config.pop("listen_elastic_ip", False):
            # The Actor.address will be announce to client, and is not on our host,
            # cannot actually listen on it,
            # so we have to keep SocketServer.host untouched to make sure Actor.address not changed
            if is_v6_ip(host):
                _host = "::"
            else:
                _host = "0.0.0.0"

        handle_channel = config.pop("handle_channel")

        # init
        UCXInitializer.init(config.get("ucx", dict()))

        async def serve_forever(client_ucp_endpoint: "ucp.Endpoint"):  # type: ignore
            try:
                await server.on_connected(
                    client_ucp_endpoint, local_address="%s:%d" % (_host, port)
                )
            except ChannelClosed:  # pragma: no cover
                logger.exception("Connection closed before handshake completed")
                return

        ucp_listener = ucp.create_listener(serve_forever, port=port)

        # get port of the ucp listener if not specified
        if not port:
            port = ucp_listener.port

        server = UCXServer(host, port, ucp_listener, channel_handler=handle_channel)
        return server

    @classmethod
    def parse_config(cls, config: dict) -> dict:
        return config

    @implements(Server.start)
    async def start(self):
        pass

    @implements(Server.join)
    async def join(self, timeout=None):
        wait_coro = self._closed.wait()
        try:
            await asyncio.wait_for(wait_coro, timeout=timeout)
        except (futures.TimeoutError, asyncio.TimeoutError):
            pass

    @implements(Server.on_connected)
    async def on_connected(self, *args, **kwargs):
        (ucp_endpoint,) = args
        local_address = kwargs.pop("local_address", None)
        dest_address = kwargs.pop("dest_address", None)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f"{type(self).__name__} got unexpected "
                f'arguments: {",".join(kwargs)}'
            )
        channel = UCXChannel(
            ucp_endpoint, local_address=local_address, dest_address=dest_address
        )
        self._channels.add(channel)
        # handle over channel to some handlers
        try:
            await self.channel_handler(channel)
        finally:
            if not channel.closed:
                await channel.close()
            # Remove channel if channel exit
            self._channels.discard(channel)
            logger.debug("Channel exit: %s", channel.info)

    @implements(Server.stop)
    async def stop(self):
        self._ucp_listener.close()
        # close all channels
        await asyncio.gather(
            *(channel.close() for channel in self._channels if not channel.closed)
        )
        self._channels.clear()
        self._ucp_listener = None
        self._closed.set()

    @property
    @implements(Server.stopped)
    def stopped(self) -> bool:
        return self._ucp_listener is None


@register_client
class UCXClient(Client):
    __slots__ = ()

    scheme = UCXServer.scheme
    channel: UCXChannel

    @classmethod
    def parse_config(cls, config: dict) -> dict:
        return config

    @staticmethod
    @implements(Client.connect)
    async def connect(
        dest_address: str, local_address: str | None = None, **kwargs
    ) -> "Client":
        prefix = f"{UCXClient.scheme}://"
        if dest_address.startswith(prefix):
            dest_address = dest_address[len(prefix) :]
        host, port_str = dest_address.rsplit(":", 1)
        port = int(port_str)
        kwargs = kwargs.copy()
        ucx_config = kwargs.pop("config", dict()).get("ucx", dict())
        UCXInitializer.init(ucx_config)

        try:
            ucp_endpoint = await ucp.create_endpoint(host, port)
        except ucp.exceptions.UCXError as e:  # pragma: no cover
            raise ChannelClosed(
                f"Connection closed before handshake completed, "
                f"local address: {local_address}, dest address: {dest_address}"
            ) from e
        channel = UCXChannel(
            ucp_endpoint, local_address=local_address, dest_address=dest_address
        )
        return UCXClient(local_address, dest_address, channel)

    async def send_buffers(self, buffers: list, meta: _MessageBase):
        return await self.channel.send_buffers(buffers, meta)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
import heapq
import json
import logging
import operator
import os
from collections import Counter
from collections.abc import Mapping

from .backends.message import SendMessage, TellMessage

logger = logging.getLogger(__name__)

XOSCAR_ENABLE_PROFILING = int(os.environ.get("XOSCAR_ENABLE_PROFILING", 0))


class _ProfilingOptionDescriptor:
    def __init__(self, _type, default):
        self._name = None
        self._type = _type
        self._default = default

    def __get__(self, obj, cls):
        if obj is None:
            return self
        v = obj._options.get(self._name)
        if v is None:
            v = os.environ.get(f"XOSCAR_PROFILING_{self._name.upper()}", self._default)
        if v is not None:
            v = self._type(v)
        # Cache the value.
        obj.__dict__[self._name] = v
        return v

    def set_name(self, name: str):
        self._name = name


class _ProfilingOptionsMeta(type):
    def __init__(cls, name, bases, classdict):
        super(_ProfilingOptionsMeta, cls).__init__(name, bases, classdict)
        for k, v in classdict.items():
            if isinstance(v, _ProfilingOptionDescriptor):
                v.set_name(k)


class _ProfilingOptions(metaclass=_ProfilingOptionsMeta):
    debug_interval_seconds = _ProfilingOptionDescriptor(float, default=None)
    slow_calls_duration_threshold = _ProfilingOptionDescriptor(int, default=1)

    def __init__(self, options):
        if isinstance(options, Mapping):
            invalid_keys = options.keys() - type(self).__dict__.keys()
            if invalid_keys:
                raise ValueError(f"Invalid profiling options: {invalid_keys}")
            self._options = options
        elif options in (True, False, None):
            self._options = {}
        else:
            raise ValueError(f"Invalid profiling options: {options}")


class DummyOperator:
    @staticmethod
    def set(key, value):
        pass

    @staticmethod
    def inc(key, value):
        pass

    @staticmethod
    def nest(key):
        return DummyOperator

    @staticmethod
    def values():
        return []

    @staticmethod
    def empty():
        return True


class ProfilingDataOperator:
    __slots__ = ("_target",)

    def __init__(self, target):
        self._target = target

    def set(self, key, value):
        self._target[key] = value

    def inc(self, key, value):
        old = self._target.get(key, 0)
        self._target[key] = old + value

    def nest(self, key):
        v = self._target.setdefault(key, {})
        if not isinstance(v, dict):
            raise TypeError(
                f"The value type of key {key} is {type(v)}, but a dict is expected."
            )
        return ProfilingDataOperator(v)

    def values(self):
        return self._target.values()

    def empty(self):
        return len(self._target) == 0


class _CallStats:
    _call_counter: Counter
    _slow_calls: list

    def __init__(self, options: _ProfilingOptions):
        self._options = options
        self._call_counter = Counter()
        self._slow_calls = []

    def collect(self, message, duration: float):
        key = (message.actor_ref.uid, message.content[0])
        self._call_counter[key] += 1
        if duration < self._options.slow_calls_duration_threshold:
            return
        slow_call_key = (
            duration,
            message.actor_ref.uid,
            message.actor_ref.address,
            message.content,
        )
        try:
            if len(self._slow_calls) < 10:
                heapq.heappush(self._slow_calls, slow_call_key)
            else:
                heapq.heapreplace(self._slow_calls, slow_call_key)
        except TypeError:
            pass

    def to_dict(self) -> dict:
        most_calls = {}
        for name_tuple, count in self._call_counter.most_common(10):
            uid, method_name = name_tuple
            most_calls[f"{uid.decode('utf-8')}.{method_name}"] = count
        slow_calls = {}
        for duration, uid, address, content in sorted(
            self._slow_calls, key=operator.itemgetter(0), reverse=True
        ):
            method_name, _batch, args, kwargs = content
            slow_calls[
                f"[{address}]{uid.decode('utf-8')}.{method_name}(args={args}, kwargs={kwargs})"
            ] = duration
        return {"most_calls": most_calls, "slow_calls": slow_calls}


class _ProfilingData:
    _instance = None

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @classmethod
    def set_instance(cls, inst: "_ProfilingData"):
        cls._instance = inst

    def __init__(self):
        self._data = {}
        self._call_stats = {}
        self._debug_task = {}

    def init(self, task_id: str, options=None):
        options = _ProfilingOptions(options)
        logger.info(
            "Init profiling data for task %s with debug interval seconds %s.",
            task_id,
            options.debug_interval_seconds,
        )
        self._data[task_id] = {
            "general": {},
            "serialization": {},
            "most_calls": {},
            "slow_calls": {},
        }
        self._call_stats[task_id] = _CallStats(options)

        async def _debug_profiling_log():
            while True:
                try:
                    r = self._data.get(task_id, None)
                    if r is None:
                        logger.info("Profiling debug log break.")
                        break
                    r = copy.copy(r)  # shadow copy is enough.
                    r.update(self._call_stats.get(task_id).to_dict())
                    logger.warning("Profiling debug:\n%s", json.dumps(r, indent=4))
                except Exception:
                    logger.exception("Profiling debug log failed.")
                await asyncio.sleep(options.debug_interval_seconds)

        if options.debug_interval_seconds is not None:
            self._debug_task[task_id] = task = asyncio.create_task(
                _debug_profiling_log()
            )
            task.add_done_callback(lambda _: self._debug_task.pop(task_id, None))

    def pop(self, task_id: str):
        logger.info("Pop profiling data of task %s.", task_id)
        debug_task = self._debug_task.pop(task_id, None)
        if debug_task is not None:
            debug_task.cancel()
        r = self._data.pop(task_id, None)
        if r is not None:
            r.update(self._call_stats.pop(task_id).to_dict())
        return r

    def collect_actor_call(self, message, duration: float):
        if self._call_stats:
            message_type = type(message)
            if message_type is SendMessage or message_type is TellMessage:
                for stats in self._call_stats.values():
                    stats.collect(message, duration)

    def __getitem__(self, item):
        key = item if isinstance(item, tuple) else (item,)
        v = None
        d = self._data
        for k in key:
            v = d.get(k, None)
            if v is None:
                break
            else:
                d = v
        return DummyOperator if v is None else ProfilingDataOperator(v)


def get_profiling_data() -> _ProfilingData:
    return _ProfilingData.get_instance()


ProfilingData = get_profiling_data()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pickle  # nosec  # pylint: disable=import_pickle

from .core import Serializer, buffered, pickle_buffers, unpickle_buffers


class UnpickleableError(Exception):
    def __init__(self, raw_error: str | Exception):
        if isinstance(raw_error, str):
            super().__init__(raw_error)
        else:
            super().__init__(
                f"Error cannot be pickled, "
                f"error type: {type(raw_error)}, "
                f"raw error:\n{raw_error}"
            )


class ExceptionSerializer(Serializer):
    @buffered
    def serial(self, obj: Exception, context: dict):
        try:
            buffers = pickle_buffers(obj)
        except (TypeError, pickle.PicklingError):
            buffers = pickle_buffers(UnpickleableError(obj))
        return (), buffers, True

    def deserial(self, serialized: tuple, context: dict, subs: list):
        return unpickle_buffers(subs)


ExceptionSerializer.register(Exception)



# Copyright 2022-2025 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import importlib
from abc import ABC, abstractmethod
from pathlib import Path


class VirtualEnvManager(ABC):
    @classmethod
    @abstractmethod
    def is_available(cls):
        pass

    def __init__(self, env_path: Path):
        self.env_path = env_path.resolve()

    @abstractmethod
    def create_env(self, python_path: Path | None = None) -> None:
        pass

    @abstractmethod
    def install_packages(self, packages: list[str], **kwargs):
        pass

    @staticmethod
    def process_packages(packages: list[str]) -> list[str]:
        """
        Process a list of package names, replacing placeholders like #system_<package>#
        with the installed version of the corresponding package from the system environment.

        Example:
            "#system_torch#" -> "torch==2.1.0" (if torch 2.1.0 is installed)

        Args:
            packages (list[str]): A list of package names, which may include placeholders.

        Returns:
            list[str]: A new list with resolved package names and versions.

        Raises:
            RuntimeError: If a specified system package is not found in the environment.
        """
        processed = []

        for pkg in packages:
            if pkg.startswith("#system_") and pkg.endswith("#"):
                real_pkg = pkg[
                    len("#system_") : -1
                ]  # Extract actual package name, e.g., "torch"
                try:
                    version = importlib.metadata.version(real_pkg)
                except importlib.metadata.PackageNotFoundError:
                    raise RuntimeError(
                        f"System package '{real_pkg}' not found. Cannot resolve '{pkg}'."
                    )
                processed.append(f"{real_pkg}=={version}")
            else:
                processed.append(pkg)

        return processed

    @abstractmethod
    def cancel_install(self):
        pass

    @abstractmethod
    def get_python_path(self) -> str | None:
        pass

    @abstractmethod
    def get_lib_path(self) -> str:
        pass

    @abstractmethod
    def remove_env(self):
        pass



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import os
from collections import defaultdict
from typing import Any, Dict, List, Optional

from .. import Actor, actor_ref
from ..context import get_context
from ..utils import lazy_import
from .common import (
    COLLECTIVE_DEVICE_ID_ENV_KEY,
    INVOKE_ERROR_MESSAGE,
    RANK_ADDRESS_ENV_KEY,
    AllReduceAlgorithm,
    CollectiveReduceOp,
)
from .process_group import ProcessGroup, ProcessGroupGloo, ProcessGroupNCCL
from .utils import get_rank_address_via_env

cupy = lazy_import("cupy")


class RankActor(Actor):
    def __init__(
        self,
        rank: int,
        world: int,
        backend: str = "gloo",
        device_id: Optional[int] = None,
        pg_options: Optional[ProcessGroup.Options] = None,
    ):
        assert backend == "gloo" or (
            backend == "nccl" and cupy is not None
        ), "cupy is required when using nccl as backend."
        self._rank = rank
        self._device_id = device_id
        self._world = world
        self._backend = backend
        self.name_to_pg: Dict[str, Dict[str, "ProcessGroup"]] = defaultdict(dict)
        self._pg_options = pg_options

    @classmethod
    def default_uid(cls):
        return "RankActor"

    async def __post_create__(self):
        os.environ[RANK_ADDRESS_ENV_KEY] = self.address
        _ip = self._get_ip()
        if self._backend == "gloo":
            pg = ProcessGroupGloo(
                _ip,
                self._rank,
                self._world,
                group_name="default",
                pg_options=self._pg_options,
            )
            self.name_to_pg["gloo"]["default"] = pg
        elif self._backend == "nccl":
            pg = ProcessGroupNCCL(
                _ip,
                self._rank,
                self._device_id,
                self._world,
                pg_options=self._pg_options,
            )
            self.name_to_pg["nccl"]["default"] = pg
        else:
            raise NotImplementedError("Not impl other backends for now!")

    def process_group(self, pg_name: str) -> ProcessGroup:
        return self.name_to_pg[self._backend][pg_name]

    def rank(self) -> int:
        return self._rank

    def world(self) -> int:
        return self._world

    def device_id(self):
        return self._device_id

    def backend(self) -> str:
        return self._backend

    def _get_ip(self) -> str:
        return self.address.rsplit(":", 1)[0]

    def _process_group_name(self, ranks: List[int]) -> str:
        return hashlib.sha1(
            bytes(self._backend + "_".join(map(str, ranks)), "utf-8")
        ).hexdigest()

    def new_group(
        self,
        ranks: List[int],
        pg_options: Optional[ProcessGroup.Options] = None,
    ) -> Optional[str]:
        assert (
            len(ranks) <= self._world
        ), "``ranks`` in new_group cannot be larger than the world."
        assert all(
            [self._world > rank >= 0 for rank in ranks]
        ), "rank in ``ranks`` is illegal."
        assert len({rank for rank in ranks}) == len(
            ranks
        ), "there can be no duplicate ranks in the ``ranks``."
        if self._rank not in ranks:
            return None
        if len(ranks) == self._world:
            return "default"
        global_ranks = sorted(ranks)
        group_rank = global_ranks.index(self._rank)
        group_world = len(global_ranks)
        group_name = self._process_group_name(global_ranks)
        device_id = self._device_id
        if group_name in self.name_to_pg[self._backend]:
            return group_name
        _ip = self._get_ip()
        if self._backend == "gloo":
            pg_gloo = ProcessGroupGloo(
                _ip,
                group_rank,
                group_world,
                group_name=group_name,
                pg_options=pg_options,
            )
            self.name_to_pg[self._backend][group_name] = pg_gloo
        elif self._backend == "nccl":
            pg_nccl = ProcessGroupNCCL(
                _ip,
                group_rank,
                device_id,  # type: ignore
                group_world,
                group_name=group_name,
                pg_options=pg_options,
            )
            self.name_to_pg[self._backend][group_name] = pg_nccl
        else:
            raise NotImplementedError("Not impl other backends for now!")
        return group_name

    def reduce(
        self,
        send_data: Any,
        recv_data: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].reduce(
                send_data, recv_data, op=op, root=root, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].reduce(
                send_data,
                recv_data,
                op=op,
                root=root,
                stream=stream,
            )

    def allreduce(
        self,
        send_data: Any,
        recv_data: Any,
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        algorithm: AllReduceAlgorithm = AllReduceAlgorithm.RING,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].allreduce(
                send_data, recv_data, op=op, algorithm=algorithm, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].allreduce(
                send_data, recv_data, op=op, stream=stream
            )

    def gather(
        self,
        send_data: Any,
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].gather(
                send_data, recv_data, root=root, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].gather(
                send_data, recv_data, root=root, stream=stream
            )

    def allgather(
        self,
        send_data: Any,
        recv_data: Any,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].allgather(
                send_data, recv_data, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].allgather(
                send_data, recv_data, stream=stream
            )

    def scatter(
        self,
        send_data: List[Any],
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].scatter(
                send_data, recv_data, root=root, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].scatter(
                send_data, recv_data, root=root, stream=stream
            )

    def reduce_scatter(
        self,
        send_data: Any,
        recv_data: Any,
        recv_elems: List[int],
        op: CollectiveReduceOp = CollectiveReduceOp.SUM,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].reduce_scatter(
                send_data, recv_data, recv_elems, op
            )
        else:
            self.name_to_pg[self._backend][pg_name].reduce_scatter(
                send_data, recv_data, recv_elems, op, stream=stream
            )

    def alltoall(
        self,
        send_data: Any,
        recv_data: Any,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].alltoall(
                send_data, recv_data, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].alltoall(
                send_data, recv_data, stream=stream
            )

    def broadcast(
        self,
        send_data: Any,
        recv_data: Any,
        root: Optional[int] = 0,
        tag: Optional[int] = 0,
        pg_name: str = "default",
        stream: Optional[Any] = None,
    ):
        assert self.backend() == "nccl" or (
            self.backend() == "gloo" and stream is None
        ), "The parameter 'stream' can only be used when the backend of the group is 'nccl'"

        if self._backend == "gloo":
            self.name_to_pg[self._backend][pg_name].broadcast(
                send_data, recv_data, root, tag=tag
            )
        else:
            self.name_to_pg[self._backend][pg_name].broadcast(
                send_data, recv_data, root, stream=stream
            )


async def init_process_group(
    rank: int,
    world_size: int,
    backend: str = "gloo",
    device_id: Optional[int] = None,
    address: Optional[str] = None,
):
    """
    Initializes the default distributed process group, and this will also
    initialize the distributed package.

    Args:
        rank (int): Rank of the current process (it should be a
                              number between 0 and ``world_size``-1).

        world_size (int): Number of processes participating in
                                    the job.

        backend (str optional): The backend to use. Depending on
                        build-time configurations, valid values include  ``gloo`` and
                        ``nccl``. If the backend is not provided, then  a ``gloo`` backend
                        will be created.

        device_id(int, optional): GPU ID the actor will bind, default ``None``
        If it is None and backend is gloo, it will try to get it from the environment variable COLLECTIVE_DEVICE_ID_ENV_KEY.
        If the environment variable is not set either, it will return an error.

        address(str, optional): actor address. default ``None``
    """
    env_device_id = os.environ.get(COLLECTIVE_DEVICE_ID_ENV_KEY, None)
    assert backend == "gloo" or (
        backend == "nccl"
        and (
            device_id is not None
            and device_id >= 0
            or env_device_id is not None
            and int(env_device_id) >= 0
        )
    ), "The device id should be set when using nccl as backend."
    assert backend == "gloo" or (
        backend == "nccl" and cupy is not None
    ), "cupy is required when using nccl as backend."
    address = address or os.environ.get(RANK_ADDRESS_ENV_KEY, None)
    if address is None:
        raise RuntimeError(
            "Cannot decide which process to involve in the collective communication."
        )
    ctx = get_context()
    if backend == "nccl" and device_id is None and env_device_id is not None:
        device_id = int(env_device_id)
    await ctx.create_actor(
        RankActor,
        rank,
        world_size,
        backend=backend,
        device_id=device_id,
        address=address,
        uid="RankActor",
    )


async def new_group(
    ranks: List[int],
    pg_options: Optional[ProcessGroup.Options] = None,
):
    """
    Creates a new distributed group.

    This function requires that all processes in the main group (i.e. all
    processes that are part of the distributed job) enter this function, even
    if they are not going to be members of the group. Additionally, groups
    should be created in the same order in all processes.

    Args:
        ranks (list[int]): List of ranks of group members. If ``None``, will be
            set to all ranks. Default is ``None``.

        pg_options (ProcessGroupOptions, optional): process group options
            specifying what additional options need to be passed in during
            the construction of specific process groups.

    Returns:
        A handle of distributed group that can be given to collective calls.
    """
    address = os.environ.get(RANK_ADDRESS_ENV_KEY, None)
    if address is None:
        raise RuntimeError(INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    return await ref.new_group(ranks, pg_options)


async def reduce(
    send_data: Any,
    recv_data: Any,
    op: CollectiveReduceOp = CollectiveReduceOp.SUM,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Reduces the numpy or cupy data across all machines.

    Only the process with rank ``root`` is going to receive the final result.

    Args:
        send_data (Any): Input of the collective. The function
            operates in-place.

        recv_data (Any): Output of the collective. The function
            operates in-place.

        root (int): Destination rank

        op (xoscar.collective.common.CollectiveReduceOp): One of the values from
            ``xoscar.collective.common.CollectiveReduceOp``
            enum.  Specifies an operation used for element-wise reductions.
            Default is ``xoscar.collective.common.CollectiveReduceOp.SUM``.

        tag (int optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.reduce(
        send_data,
        recv_data,
        op=op,
        root=root,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def allreduce(
    send_data: Any,
    recv_data: Any,
    op: CollectiveReduceOp = CollectiveReduceOp.SUM,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Reduces the numpy or cupy data across all machines in such a way that all get
    the final result.

    Args:
        send_data (Any): Input of the collective. The function
            operates in-place.

        recv_data (Any): Output of the collective. The function
            operates in-place.

        op (xoscar.collective.common.CollectiveReduceOp): One of the values from
            ``xoscar.collective.common.CollectiveReduceOp``
            enum.  Specifies an operation used for element-wise reductions.
            Default is ``xoscar.collective.common.CollectiveReduceOp.SUM``.

        algorithm (xoscar.collective.common.AllReduceAlgorithm): One of the values from
            ``xoscar.collective.common.AllReduceAlgorithm``
            enum.  Specifies an algorithm used for element-wise reductions.
            Default is ``xoscar.collective.common.AllReduceAlgorithm.RING``.

        tag (int optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid="RankActor")
    await ref.allreduce(
        send_data,
        recv_data,
        op=op,
        algorithm=AllReduceAlgorithm.RING,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def gather(
    send_data: Any,
    recv_data: Any,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Gathers a list of numpy or cupy data in a single process.

    Args:
        send_data (Any): Input data.

        recv_data (Any): Output data.

        root (int, optional): Destination rank. Default is 0.

        tag (int optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.gather(
        send_data,
        recv_data,
        root=root,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def allgather(
    send_data: Any,
    recv_data: Any,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Gathers a list of numpy or cupy data to all devices.

    Args:
        send_data (Any): Input data.

        recv_data (Any): Output data.

        tag (int optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.allgather(
        send_data,
        recv_data,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def scatter(
    send_data: List[Any],
    recv_data: Any,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Scatters a list of numpy or cupy data to all processes in a group.

    Each process will receive exactly one tensor and store its data in the
    recv_data.

    Args:
        send_data (List(Any)): Input data.

        recv_data (Any): Output data.

        root (int, optional): Source rank (default is 0).

        tag (int optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.scatter(
        send_data,
        recv_data,
        root=root,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def reduce_scatter(
    send_data: Any,
    recv_data: Any,
    recv_elems: List[int],
    op: CollectiveReduceOp = CollectiveReduceOp.SUM,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Reduces, then scatters a list of numpy or cupy data to all processes in a group.

    Args:
        send_data (Any): Input data.

        recv_data (Any): Output data.

        recv_elems (List[int]): the size of recv data for each process

        op (xoscar.collective.common.CollectiveReduceOp): One of the values from
            ``xoscar.collective.common.CollectiveReduceOp``
            enum.  Specifies an operation used for element-wise reductions.
            Default is ``xoscar.collective.common.CollectiveReduceOp.SUM``.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.reduce_scatter(
        send_data,
        recv_data,
        recv_elems,
        op,
        pg_name=group_name,
        stream=stream,
    )


async def alltoall(
    send_data: Any,
    recv_data: Any,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Each process scatters list of numpy or cupy data to all processes in a group

    Complex tensors are supported.

    Args:
        send_data (Any): Input data.

        recv_data (Any): Output data.

        tag (int, optional): Tag for this operation. default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.alltoall(
        send_data,
        recv_data,
        tag=tag,
        pg_name=group_name,
        stream=stream,
    )


async def broadcast(
    send_data: Any,
    recv_data: Any,
    root: Optional[int] = 0,
    tag: Optional[int] = 0,
    group_name: str = "default",
    stream: Optional[Any] = None,
):
    """
    Broadcasts the tensor to the whole group.

    data must have the same number of elements in all processes
    participating in the collective.

    Args:
        send_data (Any): Input data.

        recv_data (Any): Output data.

        root (int, optional): Source rank. Default is 0.

        tag (int, optional): Tag for this operation. Default is 0.

        group_name (str): The process group to work on. If None,
            the default process group will be used.

        stream (cupy.cuda.Stream, optional): stream handle for nccl, default is None.
    """
    address = get_rank_address_via_env(RANK_ADDRESS_ENV_KEY, INVOKE_ERROR_MESSAGE)
    ref = await actor_ref(address=address, uid=f"RankActor")
    await ref.broadcast(
        send_data,
        recv_data,
        root,
        tag,
        pg_name=group_name,
        stream=stream,
    )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import dataclasses
import functools
import importlib
import importlib.util as importlib_utils
import inspect
import io
import logging
import os
import random
import socket
import sys
import time
import uuid
from abc import ABC
from functools import lru_cache
from types import TracebackType
from typing import Callable, Type, Union

from ._utils import (  # noqa: F401 # pylint: disable=unused-import
    NamedType,
    Timer,
    TypeDispatcher,
    to_binary,
    to_str,
)

# Please refer to https://bugs.python.org/issue41451
try:

    class _Dummy(ABC):
        __slots__ = ("__weakref__",)

    abc_type_require_weakref_slot = True
except TypeError:
    abc_type_require_weakref_slot = False


logger = logging.getLogger(__name__)


_memory_size_indices = {"": 0, "k": 1, "m": 2, "g": 3, "t": 4}


def parse_readable_size(value: str | int | float) -> tuple[float, bool]:
    if isinstance(value, (int, float)):
        return float(value), False

    value = value.strip().lower()
    num_pos = 0
    while num_pos < len(value) and value[num_pos] in "0123456789.-":
        num_pos += 1

    value, suffix = value[:num_pos], value[num_pos:]
    suffix = suffix.strip()
    if suffix.endswith("%"):
        return float(value) / 100, True

    try:
        return float(value) * (1024 ** _memory_size_indices[suffix[:1]]), False
    except (ValueError, KeyError):
        raise ValueError(f"Unknown limitation value: {value}")


def wrap_exception(
    exc: BaseException,
    bases: tuple[Type] | tuple | None = None,
    wrap_name: str | None = None,
    message: str | None = None,
    traceback: TracebackType | None = None,
    attr_dict: dict | None = None,
) -> BaseException:
    """Generate an exception wraps the cause exception."""

    def __init__(self, *args, **kwargs):
        pass

    def __getattr__(self, item):
        return getattr(exc, item)

    def __str__(self):
        return message or super(type(self), self).__str__()

    traceback = traceback or exc.__traceback__
    bases = bases or ()
    attr_dict = attr_dict or {}
    attr_dict.update(
        {
            "__init__": __init__,
            "__getattr__": __getattr__,
            "__str__": __str__,
            "__wrapname__": wrap_name,
            "__wrapped__": exc,
            "__module__": type(exc).__module__,
            "__cause__": exc.__cause__,
            "__context__": exc.__context__,
            "__suppress_context__": exc.__suppress_context__,
            "args": exc.args,
        }
    )
    new_exc_type = type(type(exc).__name__, bases + (type(exc),), attr_dict)
    return new_exc_type().with_traceback(traceback)


# from https://github.com/ericvsmith/dataclasses/blob/master/dataclass_tools.py
# released under Apache License 2.0
def dataslots(cls):
    # Need to create a new class, since we can't set __slots__
    #  after a class has been created.

    # Make sure __slots__ isn't already set.
    if "__slots__" in cls.__dict__:  # pragma: no cover
        raise TypeError(f"{cls.__name__} already specifies __slots__")

    # Create a new dict for our new class.
    cls_dict = dict(cls.__dict__)
    field_names = tuple(f.name for f in dataclasses.fields(cls))
    cls_dict["__slots__"] = field_names
    for field_name in field_names:
        # Remove our attributes, if present. They'll still be
        #  available in _MARKER.
        cls_dict.pop(field_name, None)
    # Remove __dict__ itself.
    cls_dict.pop("__dict__", None)
    # And finally create the class.
    qualname = getattr(cls, "__qualname__", None)
    cls = type(cls)(cls.__name__, cls.__bases__, cls_dict)
    if qualname is not None:
        cls.__qualname__ = qualname
    return cls


def implements(f: Callable):
    def decorator(g):
        g.__doc__ = f.__doc__
        return g

    return decorator


class classproperty:
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


LOW_PORT_BOUND = 10000
HIGH_PORT_BOUND = 65535
_local_occupied_ports: set = set()


def _get_ports_from_netstat() -> set[int]:
    import subprocess

    while True:
        p = subprocess.Popen("netstat -a -n -p tcp".split(), stdout=subprocess.PIPE)
        try:
            outs, _ = p.communicate(timeout=5)
            lines = outs.split(to_binary(os.linesep))
            occupied = set()
            for line in lines:
                if b"." not in line:
                    continue
                line_str: str = to_str(line)
                for part in line_str.split():
                    # in windows, netstat uses ':' to separate host and port
                    part = part.replace(":", ".")
                    if "." in part:
                        _, port_str = part.rsplit(".", 1)
                        if port_str == "*":
                            continue
                        port = int(port_str)
                        if LOW_PORT_BOUND <= port <= HIGH_PORT_BOUND:
                            occupied.add(int(port_str))
                        break
            return occupied
        except subprocess.TimeoutExpired:
            p.kill()
            continue


def get_next_port(typ: int | None = None, occupy: bool = True) -> int:
    import psutil

    if sys.platform.lower().startswith("win"):
        occupied = _get_ports_from_netstat()
    else:
        try:
            conns = psutil.net_connections()
            typ = typ or socket.SOCK_STREAM
            occupied = set(
                sc.laddr.port
                for sc in conns
                if sc.type == typ and LOW_PORT_BOUND <= sc.laddr.port <= HIGH_PORT_BOUND
            )
        except psutil.AccessDenied:
            occupied = _get_ports_from_netstat()

    occupied.update(_local_occupied_ports)
    random.seed(uuid.uuid1().bytes)
    randn = random.randint(0, 100000000)

    idx = int(randn % (1 + HIGH_PORT_BOUND - LOW_PORT_BOUND - len(occupied)))
    for i in range(LOW_PORT_BOUND, HIGH_PORT_BOUND + 1):
        if i in occupied:
            continue
        if idx == 0:
            if occupy:
                _local_occupied_ports.add(i)
            return i
        idx -= 1
    raise SystemError("No ports available.")


def lazy_import(
    name: str,
    package: str | None = None,
    globals: dict | None = None,  # pylint: disable=redefined-builtin
    locals: dict | None = None,  # pylint: disable=redefined-builtin
    rename: str | None = None,
    placeholder: bool = False,
):
    rename = rename or name
    prefix_name = name.split(".", 1)[0]
    globals = globals or inspect.currentframe().f_back.f_globals  # type: ignore

    class LazyModule:
        def __init__(self):
            self._on_loads = []

        def __getattr__(self, item):
            if item.startswith("_pytest") or item in ("__bases__", "__test__"):
                raise AttributeError(item)

            real_mod = importlib.import_module(name, package=package)
            if rename in globals:
                globals[rename] = real_mod
            elif locals is not None:
                locals[rename] = real_mod
            ret = getattr(real_mod, item)
            for on_load_func in self._on_loads:
                on_load_func()
            # make sure on_load hooks only executed once
            self._on_loads = []
            return ret

        def add_load_handler(self, func: Callable):
            self._on_loads.append(func)
            return func

    if importlib_utils.find_spec(prefix_name) is not None:
        return LazyModule()
    elif placeholder:
        return ModulePlaceholder(prefix_name)
    else:
        return None


def lazy_import_on_load(lazy_mod):
    def wrapper(fun):
        if lazy_mod is not None and hasattr(lazy_mod, "add_load_handler"):
            lazy_mod.add_load_handler(fun)
        return fun

    return wrapper


class ModulePlaceholder:
    def __init__(self, mod_name: str):
        self._mod_name = mod_name

    def _raises(self):
        raise AttributeError(f"{self._mod_name} is required but not installed.")

    def __getattr__(self, key):
        self._raises()

    def __call__(self, *_args, **_kwargs):
        self._raises()


def patch_asyncio_task_create_time():  # pragma: no cover
    new_loop = False
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        new_loop = True
    loop_class = loop.__class__
    # Save raw loop_class.create_task and make multiple apply idempotent
    loop_create_task = getattr(
        patch_asyncio_task_create_time, "loop_create_task", loop_class.create_task
    )
    patch_asyncio_task_create_time.loop_create_task = loop_create_task

    def new_loop_create_task(*args, **kwargs):
        task = loop_create_task(*args, **kwargs)
        task.__xoscar_asyncio_task_create_time__ = time.time()
        return task

    if loop_create_task is not new_loop_create_task:
        loop_class.create_task = new_loop_create_task
    if not new_loop and loop.create_task is not new_loop_create_task:
        loop.create_task = functools.partial(new_loop_create_task, loop)


async def asyncio_task_timeout_detector(
    check_interval: int, task_timeout_seconds: int, task_exclude_filters: list[str]
):
    task_exclude_filters.append("asyncio_task_timeout_detector")
    while True:  # pragma: no cover
        await asyncio.sleep(check_interval)
        loop = asyncio.get_running_loop()
        current_time = (
            time.time()
        )  # avoid invoke `time.time()` frequently if we have plenty of unfinished tasks.
        for task in asyncio.all_tasks(loop=loop):
            # Some task may be create before `patch_asyncio_task_create_time` applied, take them as never timeout.
            create_time = getattr(
                task, "__xoscar_asyncio_task_create_time__", current_time
            )
            if current_time - create_time >= task_timeout_seconds:
                stack = io.StringIO()
                task.print_stack(file=stack)
                task_str = str(task)
                if any(
                    excluded_task in task_str for excluded_task in task_exclude_filters
                ):
                    continue
                logger.warning(
                    """Task %s in event loop %s doesn't finish in %s seconds. %s""",
                    task,
                    loop,
                    time.time() - create_time,
                    stack.getvalue(),
                )


def register_asyncio_task_timeout_detector(
    check_interval: int | None = None,
    task_timeout_seconds: int | None = None,
    task_exclude_filters: list[str] | None = None,
) -> asyncio.Task | None:  # pragma: no cover
    """Register a asyncio task which print timeout task periodically."""
    check_interval = check_interval or int(
        os.environ.get("XOSCAR_DEBUG_ASYNCIO_TASK_TIMEOUT_CHECK_INTERVAL", -1)
    )
    if check_interval > 0:
        patch_asyncio_task_create_time()
        task_timeout_seconds = task_timeout_seconds or int(
            os.environ.get("XOSCAR_DEBUG_ASYNCIO_TASK_TIMEOUT_SECONDS", check_interval)
        )
        if not task_exclude_filters:
            # Ignore Xoscar by default since it has some long-running coroutines.
            task_exclude_filter = os.environ.get(
                "XOSCAR_DEBUG_ASYNCIO_TASK_EXCLUDE_FILTERS", "xoscar"
            )
            task_exclude_filters = task_exclude_filter.split(";")
        if sys.version_info[:2] < (3, 7):
            logger.warning(
                "asyncio tasks timeout detector is not supported under python %s",
                sys.version,
            )
        else:
            loop = asyncio.get_running_loop()
            logger.info(
                "Create asyncio tasks timeout detector with check_interval %s task_timeout_seconds %s "
                "task_exclude_filters %s",
                check_interval,
                task_timeout_seconds,
                task_exclude_filters,
            )
            return loop.create_task(
                asyncio_task_timeout_detector(
                    check_interval, task_timeout_seconds, task_exclude_filters
                )
            )
    else:
        return None


def ensure_coverage():
    # make sure coverage is handled when starting with subprocess.Popen
    if (
        not sys.platform.startswith("win") and "COV_CORE_SOURCE" in os.environ
    ):  # pragma: no cover
        try:
            from pytest_cov.embed import cleanup_on_sigterm
        except ImportError:
            pass
        else:
            cleanup_on_sigterm()


def retry_callable(
    callable_,
    ex_type: type = Exception,
    wait_interval=1,
    max_retries=-1,
    sync: bool | None = None,
):
    if inspect.iscoroutinefunction(callable_) or sync is False:

        @functools.wraps(callable)
        async def retry_call(*args, **kwargs):
            num_retried = 0
            while max_retries < 0 or num_retried < max_retries:
                num_retried += 1
                try:
                    return await callable_(*args, **kwargs)
                except ex_type:
                    await asyncio.sleep(wait_interval)

    else:

        @functools.wraps(callable)
        def retry_call(*args, **kwargs):
            num_retried = 0
            ex = None
            while max_retries < 0 or num_retried < max_retries:
                num_retried += 1
                try:
                    return callable_(*args, **kwargs)
                except ex_type as e:
                    ex = e
                    time.sleep(wait_interval)
            assert ex is not None
            raise ex  # pylint: disable-msg=E0702

    return retry_call


_cupy = lazy_import("cupy")
_rmm = lazy_import("rmm")


def is_cuda_buffer(cuda_buffer: Union["_cupy.ndarray", "_rmm.DeviceBuffer"]) -> bool:  # type: ignore
    return hasattr(cuda_buffer, "__cuda_array_interface__")


def is_windows():
    return sys.platform.startswith("win")


def is_linux():
    return sys.platform.startswith("linux")


@lru_cache
def is_py_312():
    return sys.version_info[:2] == (3, 12)


def is_v4_zero_ip(ip_port_addr: str) -> bool:
    return ip_port_addr.split("://")[-1].startswith("0.0.0.0:")


def is_v6_zero_ip(ip_port_addr: str) -> bool:
    # tcp6 addr ":::123", ":: means all zero"
    arr = ip_port_addr.split("://")[-1].split(":")
    if len(arr) <= 2:  # Not tcp6 or udp6
        return False
    for part in arr[0:-1]:
        if part != "":
            if int(part, 16) != 0:
                return False
    return True


def is_zero_ip(ip_port_addr: str) -> bool:
    return is_v4_zero_ip(ip_port_addr) or is_v6_zero_ip(ip_port_addr)


def is_v6_ip(ip_port_addr: str) -> bool:
    arr = ip_port_addr.split("://", 1)[-1].split(":")
    return len(arr) > 1


def fix_all_zero_ip(remote_addr: str, connect_addr: str) -> str:
    """
    Use connect_addr to fix ActorRef.address return by remote server.
    When remote server listen on "0.0.0.0:port" or ":::port", it will return ActorRef.address set to listening addr,
    it cannot be use by client for the following interaction unless we fix it.
    (client will treat 0.0.0.0 as 127.0.0.1)

    NOTE: Server might return a different addr from a pool for load-balance purpose.
    """
    if remote_addr == connect_addr:
        return remote_addr
    if not is_v4_zero_ip(remote_addr) and not is_v6_zero_ip(remote_addr):
        # Remote server returns on non-zero ip
        return remote_addr
    if is_v4_zero_ip(connect_addr) or is_v6_zero_ip(connect_addr):
        # Client connect to local server
        return remote_addr
    remote_port = remote_addr.split(":")[-1]
    connect_ip = ":".join(connect_addr.split(":")[0:-1])  # Remote the port
    return f"{connect_ip}:{remote_port}"



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import dataclasses
import logging
import os
import subprocess
import sys
import uuid
from collections import namedtuple
from ctypes import (
    CDLL,
    POINTER,
    Structure,
    byref,
    c_char,
    c_char_p,
    c_int,
    c_uint,
    c_ulonglong,
    create_string_buffer,
)
from typing import List, Tuple, Union

from .utils import parse_readable_size

logger = logging.getLogger(__name__)

# Some constants taken from cuda.h
CUDA_SUCCESS = 0
CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16
CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39
CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13
CU_DEVICE_ATTRIBUTE_PCI_BUS_ID = 33
CU_DEVICE_ATTRIBUTE_PCI_DEVICE_ID = 34
CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36

CU_NO_CUDA_CAPABLE_DEVICE_DETECTED = 100

# nvml constants
NVML_SUCCESS = 0
NVML_ERROR_UNINITIALIZED = 1
NVML_ERROR_INVALID_ARGUMENT = 2
NVML_ERROR_NOT_SUPPORTED = 3
NVML_ERROR_NO_PERMISSION = 4
NVML_ERROR_ALREADY_INITIALIZED = 5
NVML_ERROR_NOT_FOUND = 6
NVML_ERROR_INSUFFICIENT_SIZE = 7
NVML_ERROR_INSUFFICIENT_POWER = 8
NVML_ERROR_DRIVER_NOT_LOADED = 9
NVML_ERROR_TIMEOUT = 10
NVML_ERROR_IRQ_ISSUE = 11
NVML_ERROR_LIBRARY_NOT_FOUND = 12
NVML_ERROR_FUNCTION_NOT_FOUND = 13
NVML_ERROR_CORRUPTED_INFOROM = 14
NVML_ERROR_GPU_IS_LOST = 15
NVML_ERROR_RESET_REQUIRED = 16
NVML_ERROR_OPERATING_SYSTEM = 17
NVML_ERROR_LIB_RM_VERSION_MISMATCH = 18
NVML_ERROR_IN_USE = 19
NVML_ERROR_MEMORY = 20
NVML_ERROR_NO_DATA = 21
NVML_ERROR_VGPU_ECC_NOT_SUPPORTED = 22
NVML_ERROR_INSUFFICIENT_RESOURCES = 23
NVML_ERROR_FREQ_NOT_SUPPORTED = 24
NVML_ERROR_UNKNOWN = 999
NVML_TEMPERATURE_GPU = 0
NVML_DRIVER_NOT_LOADED = 9
NVML_DEVICE_UUID_V2_BUFFER_SIZE = 96
NVML_VALUE_NOT_AVAILABLE_ulonglong = c_ulonglong(-1)
NVML_DEVICE_MIG_DISABLE = 0x0
NVML_DEVICE_MIG_ENABLE = 0x1


class _CUuuid_t(Structure):
    _fields_ = [("bytes", c_char * 16)]


class _nvmlUtilization_t(Structure):
    _fields_ = [
        ("gpu", c_uint),
        ("memory", c_uint),
    ]


class _struct_nvmlDevice_t(Structure):
    pass  # opaque handle


_nvmlDevice_t = POINTER(_struct_nvmlDevice_t)


class _nvmlBAR1Memory_t(Structure):
    _fields_ = [
        ("total", c_ulonglong),
        ("free", c_ulonglong),
        ("used", c_ulonglong),
    ]


class _nvmlProcessInfo_t(Structure):
    _fields_ = [
        ("pid", c_uint),
        ("usedGpuMemory", c_ulonglong),
        ("gpuInstanceId", c_uint),
        ("computeInstanceId", c_uint),
    ]


## Alternative object
# Allows the object to be printed
# Allows mismatched types to be assigned
#  - like None when the Structure variant requires c_uint
class nvmlFriendlyObject:
    def __init__(self, dictionary):
        for x in dictionary:
            setattr(self, x, dictionary[x])

    def __str__(self):
        return self.__dict__.__str__()


def nvmlStructToFriendlyObject(struct):
    d = {}
    for x in struct._fields_:
        key = x[0]
        value = getattr(struct, key)
        # only need to convert from bytes if bytes, no need to check python version.
        d[key] = value.decode() if isinstance(value, bytes) else value
    obj = nvmlFriendlyObject(d)
    return obj


@dataclasses.dataclass
class CudaDeviceInfo:
    uuid: bytes | None = None
    device_index: int | None = None
    mig_index: int | None = None


@dataclasses.dataclass
class CudaContext:
    has_context: bool
    device_info: CudaDeviceInfo | None = None


_is_windows: bool = sys.platform.startswith("win")
_is_wsl: bool = "WSL_DISTRO_NAME" in os.environ


def _load_nv_library(*libnames):
    for lib in libnames:
        try:
            return CDLL(lib)
        except OSError:
            continue


_cuda_lib = _nvml_lib = None

_cu_device_info = namedtuple(
    "_cu_device_info", "index uuid name multiprocessors cuda_cores threads"
)
_nvml_driver_info = namedtuple("_nvml_driver_info", "driver_version cuda_version")
_nvml_device_status = namedtuple(
    "_nvml_device_status",
    "gpu_util mem_util temperature fb_total_mem fb_used_mem fb_free_mem",
)

_init_pid = None
_gpu_count = None
_driver_info = None
_device_infos: dict[int, _cu_device_info] = dict()

_no_device_warned = False


class NVError(Exception):
    def __init__(self, msg, *args, errno=None):
        self._errno = errno
        super().__init__(msg or "Unknown error", *args)

    def __str__(self):
        return f"({self._errno}) {super().__str__()}"

    @property
    def errno(self):
        return self._errno

    @property
    def message(self):
        return super().__str__()


class NVDeviceAPIError(NVError):
    pass


class NVMLAPIError(NVError):
    pass


def _cu_check_error(result):
    if result != CUDA_SUCCESS:
        _error_str = c_char_p()
        _cuda_lib.cuGetErrorString(result, byref(_error_str))
        err_value = _error_str.value.decode() if _error_str.value is not None else None
        raise NVDeviceAPIError(err_value, errno=result)


_nvmlErrorString = None


def _nvml_check_error(result):
    global _nvmlErrorString
    if _nvmlErrorString is None:
        _nvmlErrorString = _nvml_lib.nvmlErrorString
        _nvmlErrorString.restype = c_char_p

    if result != NVML_SUCCESS:
        _error_str = _nvmlErrorString(result)
        raise NVMLAPIError(_error_str.decode(), errno=result)


_cu_process_var_to_cores = {
    (1, 0): 8,
    (1, 1): 8,
    (1, 2): 8,
    (1, 3): 8,
    (2, 0): 32,
    (2, 1): 48,
}


def _cu_get_processor_cores(major, minor):
    return _cu_process_var_to_cores.get((major, minor), 192)


def _init_cp():
    global _cuda_lib, _no_device_warned
    if _init_pid == os.getpid():
        return

    libcuda_paths = ["libcuda.so", "libcuda.dylib", "cuda.dll", "nvcuda.dll"]
    if _is_wsl:
        libcuda_paths = ["/usr/lib/wsl/lib/libcuda.so"] + libcuda_paths
    _cuda_lib = _load_nv_library(*libcuda_paths)

    if _cuda_lib is None:
        return
    try:
        _cu_check_error(_cuda_lib.cuInit(0))
    except NVDeviceAPIError as ex:
        if ex.errno == CU_NO_CUDA_CAPABLE_DEVICE_DETECTED:
            _cuda_lib = None
            if not _no_device_warned:
                logger.warning("No CUDA device detected")
                _no_device_warned = True
        else:
            logger.exception("Failed to initialize libcuda.")
        return


def _init_nvml():
    global _nvml_lib, _no_device_warned
    if _init_pid == os.getpid():
        return

    nvml_paths = [
        "libnvidia-ml.so",
        "libnvidia-ml.so.1",
        "libnvidia-ml.dylib",
        "nvml.dll",
    ]
    if _is_windows:
        nvml_paths.append(
            os.path.join(
                os.getenv("ProgramFiles", "C:/Program Files"),
                "NVIDIA Corporation/NVSMI/nvml.dll",
            )
        )
    if _is_wsl:
        nvml_paths = ["/usr/lib/wsl/lib/libnvidia-ml.so.1"] + nvml_paths
    _nvml_lib = _load_nv_library(*nvml_paths)

    if _nvml_lib is None:
        return
    try:
        _nvml_check_error(_nvml_lib.nvmlInit_v2())
    except NVMLAPIError as ex:
        if ex.errno == NVML_DRIVER_NOT_LOADED:
            _nvml_lib = None
            if not _no_device_warned:
                logger.warning(
                    "Failed to load libnvidia-ml: %s, no CUDA device will be enabled",
                    ex.message,
                )
                _no_device_warned = True
        else:
            logger.exception("Failed to initialize libnvidia-ml.")
        return


def _init():
    global _init_pid

    _init_cp()
    _init_nvml()

    if _nvml_lib is not None and _cuda_lib is not None:
        _init_pid = os.getpid()


def get_device_count() -> int | None:
    global _gpu_count

    if _gpu_count is not None:
        return _gpu_count

    _init_nvml()
    if _nvml_lib is None:
        return None

    if "CUDA_VISIBLE_DEVICES" in os.environ:
        devices = os.environ["CUDA_VISIBLE_DEVICES"].strip()
        if not devices or devices == "-1":
            _gpu_count = 0
        else:
            _gpu_count = len(devices.split(","))
    else:
        n_gpus = c_uint()
        _cu_check_error(_nvml_lib.nvmlDeviceGetCount(byref(n_gpus)))
        _gpu_count = n_gpus.value
    return _gpu_count


def _get_all_device_count() -> int | None:
    _init_nvml()
    if _nvml_lib is None:
        return None

    n_gpus = c_uint()
    _cu_check_error(_nvml_lib.nvmlDeviceGetCount(byref(n_gpus)))
    return n_gpus.value


def get_driver_info() -> _nvml_driver_info | None:
    global _driver_info

    _init_nvml()
    if _nvml_lib is None:
        return None
    if _driver_info is not None:
        return _driver_info

    version_buf = create_string_buffer(100)
    cuda_version = c_uint()

    _nvml_check_error(
        _nvml_lib.nvmlSystemGetDriverVersion(version_buf, len(version_buf))
    )
    _nvml_check_error(_nvml_lib.nvmlSystemGetCudaDriverVersion(byref(cuda_version)))

    _driver_info = _nvml_driver_info(
        driver_version=version_buf.value.decode(),
        cuda_version=".".join(str(v) for v in divmod(cuda_version.value, 1000)),
    )
    return _driver_info


def get_device_info(dev_index: int) -> _cu_device_info | None:
    try:
        return _device_infos[dev_index]
    except KeyError:
        pass

    _init()
    if _init_pid is None:
        return None

    device = c_int()
    name_buf = create_string_buffer(100)
    uuid_t = _CUuuid_t()
    cc_major = c_int()
    cc_minor = c_int()
    cores = c_int()
    threads_per_core = c_int()

    _cu_check_error(_cuda_lib.cuDeviceGet(byref(device), c_int(dev_index)))  # type: ignore
    _cu_check_error(_cuda_lib.cuDeviceGetName(name_buf, len(name_buf), device))  # type: ignore
    _cu_check_error(_cuda_lib.cuDeviceGetUuid(byref(uuid_t), device))  # type: ignore
    _cu_check_error(
        _cuda_lib.cuDeviceComputeCapability(byref(cc_major), byref(cc_minor), device)  # type: ignore
    )
    _cu_check_error(
        _cuda_lib.cuDeviceGetAttribute(  # type: ignore
            byref(cores), CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT, device
        )
    )
    _cu_check_error(
        _cuda_lib.cuDeviceGetAttribute(  # type: ignore
            byref(threads_per_core),
            CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR,
            device,
        )
    )

    if "CUDA_VISIBLE_DEVICES" in os.environ:
        real_dev_index = [
            int(s) for s in os.environ["CUDA_VISIBLE_DEVICES"].split(",")
        ][dev_index]
    else:
        real_dev_index = dev_index

    info = _device_infos[dev_index] = _cu_device_info(
        index=real_dev_index,
        uuid=uuid.UUID(bytes=uuid_t.bytes),
        name=name_buf.value.decode(),
        multiprocessors=cores.value,
        cuda_cores=cores.value
        * _cu_get_processor_cores(cc_major.value, cc_minor.value),
        threads=cores.value * threads_per_core.value,
    )
    return info


def get_device_status(dev_index: int) -> _nvml_device_status | None:
    _init()
    if _init_pid is None:
        return None

    c_device = _nvmlDevice_t()
    c_utils = _nvmlUtilization_t()
    c_temperature = c_uint()
    c_memory_info = _nvmlBAR1Memory_t()

    dev_uuid = get_device_info(dev_index).uuid  # type: ignore

    uuid_str = ("GPU-" + str(dev_uuid)).encode()

    if not _is_wsl:
        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetHandleByUUID(uuid_str, byref(c_device))  # type: ignore
        )

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetUtilizationRates(c_device, byref(c_utils))  # type: ignore
        )
        gpu_util = c_utils.gpu
        mem_util = c_utils.memory

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetTemperature(  # type: ignore
                c_device, NVML_TEMPERATURE_GPU, byref(c_temperature)
            )
        )
        temperature = c_temperature.value

        _nvml_check_error(
            _nvml_lib.nvmlDeviceGetMemoryInfo(c_device, byref(c_memory_info))  # type: ignore
        )
        fb_total_mem = c_memory_info.total
        fb_free_mem = c_memory_info.free
        fb_used_mem = c_memory_info.used
    else:
        import defusedxml

        proc = subprocess.Popen(
            ["/usr/lib/wsl/lib/nvidia-smi", "-q", f"--id={dev_index}", "-x"],
            stdout=subprocess.PIPE,
        )
        proc.wait()
        xml_result = defusedxml.ElementTree.fromstring(proc.stdout.read())  # type: ignore
        gpu_node = xml_result.find("gpu")

        fb_node = gpu_node.find("fb_memory_usage")
        fb_total_mem = int(parse_readable_size(fb_node.find("total").text)[0])
        fb_free_mem = int(parse_readable_size(fb_node.find("free").text)[0])
        fb_used_mem = int(parse_readable_size(fb_node.find("used").text)[0])

        util_node = gpu_node.find("utilization")
        if util_node.find("gpu_util").text == "N/A":
            gpu_util = 0
        else:
            gpu_util = int(util_node.find("gpu_util"))
        if util_node.find("memory_util").text == "N/A":
            mem_util = 0
        else:
            mem_util = int(util_node.find("memory_util"))

        temperature = int(gpu_node.find("temperature").find("gpu_temp").text[:-1])

    return _nvml_device_status(
        gpu_util=gpu_util,
        mem_util=mem_util,
        temperature=temperature,
        fb_total_mem=fb_total_mem,
        fb_free_mem=fb_free_mem,
        fb_used_mem=fb_used_mem,
    )


def get_handle_by_index(index: int) -> _nvmlDevice_t:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_index = c_int(index)
    device = _nvmlDevice_t()
    _nvml_check_error(_nvml_lib.nvmlDeviceGetHandleByIndex_v2(c_index, byref(device)))
    return device


def get_handle_by_uuid(uuid: bytes) -> _nvmlDevice_t:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_uuid = c_char_p(uuid)
    device = _nvmlDevice_t()
    _nvml_check_error(_nvml_lib.nvmlDeviceGetHandleByUUID(c_uuid, byref(device)))
    return device


def get_mig_mode(device: _nvmlDevice_t) -> Tuple[int, int] | None:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_current_mode, c_pending_mode = c_uint(), c_uint()
    _nvml_check_error(
        _nvml_lib.nvmlDeviceGetMigMode(
            device, byref(c_current_mode), byref(c_pending_mode)
        )
    )
    return c_current_mode.value, c_pending_mode.value


def get_max_mig_device_count(device: _nvmlDevice_t) -> int | None:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_count = c_uint()
    _nvml_check_error(_nvml_lib.nvmlDeviceGetMaxMigDeviceCount(device, byref(c_count)))
    return c_count.value


def get_mig_device_handle_by_index(device: _nvmlDevice_t, index: int) -> _nvmlDevice_t:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_index = c_uint(index)
    mig_device = _nvmlDevice_t()
    _nvml_check_error(
        _nvml_lib.nvmlDeviceGetMigDeviceHandleByIndex(
            device, c_index, byref(mig_device)
        )
    )
    return mig_device


def get_index(handle: _nvmlDevice_t) -> int | None:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_index = c_uint()
    _nvml_check_error(_nvml_lib.nvmlDeviceGetIndex(handle, byref(c_index)))
    return c_index.value


def get_uuid(handle: _nvmlDevice_t) -> bytes | None:  # type: ignore
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_uuid = create_string_buffer(NVML_DEVICE_UUID_V2_BUFFER_SIZE)
    _nvml_check_error(
        _nvml_lib.nvmlDeviceGetUUID(
            handle, c_uuid, c_uint(NVML_DEVICE_UUID_V2_BUFFER_SIZE)
        )
    )
    return c_uuid.value


def get_index_and_uuid(device: Union[int, bytes, str]) -> CudaDeviceInfo | None:
    _init_nvml()
    if _nvml_lib is None:
        return None

    try:
        device_index = int(device)
        device_handle = get_handle_by_index(device_index)
        uuid = get_uuid(device_handle)
    except ValueError:
        uuid = device if isinstance(device, bytes) else device.encode()  # type: ignore
        uuid_handle = get_handle_by_uuid(uuid)
        device_index = get_index(uuid_handle)  # type: ignore
        uuid = get_uuid(uuid_handle)

    return CudaDeviceInfo(uuid=uuid, device_index=device_index)


def get_compute_running_processes(  # type: ignore
    handle: _nvmlDevice_t,  # type: ignore
) -> List[nvmlFriendlyObject] | None:
    _init_nvml()
    if _nvml_lib is None:
        return None

    c_count = c_uint(0)
    func = getattr(_nvml_lib, "nvmlDeviceGetComputeRunningProcesses_v3", None)
    if func is None:
        func = getattr(_nvml_lib, "nvmlDeviceGetComputeRunningProcesses_v2")
    ret = func(handle, byref(c_count), None)

    if ret == NVML_SUCCESS:
        # special case, no running processes
        return []
    elif ret == NVML_ERROR_INSUFFICIENT_SIZE:
        # typical case
        # oversize the array in case more processes are created
        c_count.value = c_count.value * 2 + 5
        proc_array = _nvmlProcessInfo_t * c_count.value
        c_procs = proc_array()

        _nvml_check_error(func(handle, byref(c_count), c_procs))

        procs = []
        for i in range(c_count.value):
            # use an alternative struct for this object
            obj = nvmlStructToFriendlyObject(c_procs[i])
            if obj.usedGpuMemory == NVML_VALUE_NOT_AVAILABLE_ulonglong.value:
                # special case for WDDM on Windows, see comment above
                obj.usedGpuMemory = None
            procs.append(obj)

        return procs
    else:
        # error case
        _nvml_check_error(ret)


def _running_process_matches(handle: _nvmlDevice_t) -> bool:  # type: ignore
    """Check whether the current process is same as that of handle
    Parameters
    ----------
    handle : _nvmlDevice_t
        NVML handle to CUDA device
    Returns
    -------
    out : bool
        Whether the device handle has a CUDA context on the running process.
    """
    return any(os.getpid() == o.pid for o in get_compute_running_processes(handle))  # type: ignore


def get_cuda_context() -> CudaContext:
    """Check whether the current process already has a CUDA context created."""

    _init()
    if _init_pid is None:
        return CudaContext(has_context=False)

    for index in range(_get_all_device_count()):  # type: ignore
        handle = get_handle_by_index(index)
        try:
            mig_current_mode, mig_pending_mode = get_mig_mode(handle)  # type: ignore
        except NVMLAPIError as e:
            if e.errno == NVML_ERROR_NOT_SUPPORTED:
                mig_current_mode = NVML_DEVICE_MIG_DISABLE
            else:
                raise
        if mig_current_mode == NVML_DEVICE_MIG_ENABLE:
            for mig_index in range(get_max_mig_device_count(handle)):  # type: ignore
                try:
                    mig_handle = get_mig_device_handle_by_index(handle, mig_index)
                except NVMLAPIError as e:
                    if e.errno == NVML_ERROR_NOT_FOUND:
                        # No MIG device with that index
                        continue
                    else:
                        raise
                if _running_process_matches(mig_handle):
                    return CudaContext(
                        has_context=True,
                        device_info=CudaDeviceInfo(
                            uuid=get_uuid(handle),
                            device_index=index,
                            mig_index=mig_index,
                        ),
                    )
        else:
            if _running_process_matches(handle):
                return CudaContext(
                    has_context=True,
                    device_info=CudaDeviceInfo(
                        uuid=get_uuid(handle), device_index=index
                    ),
                )

    return CudaContext(has_context=False)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import struct
from io import BytesIO
from typing import Any, BinaryIO, Union

import cloudpickle
import numpy as np

from ..utils import lazy_import
from .core import deserialize, serialize_with_spawn

rmm = lazy_import("rmm")

DEFAULT_SERIALIZATION_VERSION = 1
DEFAULT_SPAWN_THRESHOLD = 100
BUFFER_SIZES_NAME = "buf_sizes"


class AioSerializer:
    def __init__(self, obj: Any, compress=0):
        self._obj = obj
        self._compress = compress

    async def _get_buffers(self):
        headers, buffers = await serialize_with_spawn(
            self._obj, spawn_threshold=DEFAULT_SPAWN_THRESHOLD
        )

        def _is_cuda_buffer(buf: Union["rmm.DeviceBuffer", BinaryIO]):  # type: ignore
            return hasattr(buf, "__cuda_array_interface__")

        is_cuda_buffers = [_is_cuda_buffer(buf) for buf in buffers]
        headers[0]["is_cuda_buffers"] = np.array(is_cuda_buffers)

        # add buffer lengths into headers
        headers[0][BUFFER_SIZES_NAME] = [
            buf.nbytes if hasattr(buf, "nbytes") else len(buf) for buf in buffers
        ]
        header = cloudpickle.dumps(headers)

        # gen header buffer
        header_bio = BytesIO()
        # write version first
        header_bio.write(struct.pack("B", DEFAULT_SERIALIZATION_VERSION))
        # write header length
        header_bio.write(struct.pack("<Q", len(header)))
        # write compression
        header_bio.write(struct.pack("<H", self._compress))

        out_buffers = [header_bio.getbuffer(), header]
        out_buffers.extend(buffers)

        return out_buffers

    async def run(self):
        return await self._get_buffers()


MALFORMED_MSG = "Received malformed data, please check Xoscar version on both side."


def get_header_length(header_bytes: bytes):
    version = struct.unpack("B", header_bytes[:1])[0]
    # now we only have default version
    if version != DEFAULT_SERIALIZATION_VERSION:
        # when version not matched,
        # we will immediately abort the connection
        # EOFError will be captured by channel
        raise EOFError(MALFORMED_MSG)
    # header length
    header_length = struct.unpack("<Q", header_bytes[1:9])[0]
    # compress
    _ = struct.unpack("<H", header_bytes[9:])[0]
    return header_length


class AioDeserializer:
    def __init__(self, file):
        self._file = file

    def _readexactly(self, n: int):
        # asyncio StreamReader may not guarantee to read n bytes
        # for it we need to call `readexactly` instead
        read = (
            self._file.readexactly
            if hasattr(self._file, "readexactly")
            else self._file.read
        )
        return read(n)

    async def _get_obj_header_bytes(self):
        try:
            header_bytes = bytes(await self._readexactly(11))
        except ConnectionResetError:
            raise EOFError("Server may be closed")
        if len(header_bytes) == 0:
            raise EOFError("Received empty bytes")
        header_length = get_header_length(header_bytes)
        return await self._readexactly(header_length)

    async def _get_obj(self):
        header = cloudpickle.loads(await self._get_obj_header_bytes())
        # get buffer size
        buffer_sizes = header[0].pop(BUFFER_SIZES_NAME)
        # get buffers
        buffers = [await self._readexactly(size) for size in buffer_sizes]
        # get num of objs
        num_objs = header[0].get("_N", 0)

        if num_objs <= DEFAULT_SPAWN_THRESHOLD:
            return deserialize(header, buffers)
        else:
            return await asyncio.to_thread(deserialize, header, buffers)

    async def run(self):
        return await self._get_obj()

    async def get_size(self):
        # extract header
        header_bytes = await self._get_obj_header_bytes()
        header = cloudpickle.loads(header_bytes)
        # get buffer size
        buffer_sizes = header[0].pop(BUFFER_SIZES_NAME)
        return 11 + len(header_bytes) + sum(buffer_sizes)

    async def get_header(self):
        return cloudpickle.loads(await self._get_obj_header_bytes())



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple

import numpy as np

try:
    import scipy.sparse as sps
except ImportError:  # pragma: no cover
    sps = None

from .core import Serializer, buffered, deserialize, serialize


class CsrMatrixSerializer(Serializer):
    @buffered
    def serial(self, obj: Any, context: Dict):
        data_header, data_buffers = serialize(obj.data)
        idx_header, idx_buffers = serialize(obj.indices)
        indptr_header, indptr_buffers = serialize(obj.indptr)
        header = (
            data_header,  # data_header
            len(data_buffers),  # data_buf_num
            idx_header,  # idx_header
            len(idx_buffers),  # idx_buf_num
            indptr_header,  # indptr_header
            obj.shape,  # shape
        )
        return header, data_buffers + idx_buffers + indptr_buffers, True

    def deserial(self, serialized: Tuple, context: Dict, subs: List):
        (
            data_header,
            data_buf_num,
            idx_header,
            idx_buf_num,
            indptr_header,
            shape,
        ) = serialized
        data_buffers = subs[:data_buf_num]
        idx_buffers = subs[data_buf_num : data_buf_num + idx_buf_num]
        indptr_buffers = subs[data_buf_num + idx_buf_num :]

        data = deserialize(data_header, data_buffers)
        indices = deserialize(idx_header, idx_buffers)
        indptr = deserialize(indptr_header, indptr_buffers)
        shape = tuple(shape)

        empty_arr = np.zeros(0, dtype=data.dtype)

        target_csr = sps.coo_matrix(
            (empty_arr, (empty_arr,) * 2), dtype=data.dtype, shape=shape
        ).tocsr()
        target_csr.data, target_csr.indices, target_csr.indptr = data, indices, indptr
        return target_csr


if sps:  # pragma: no branch
    CsrMatrixSerializer.register(sps.csr_matrix)



# distutils: language = c++
# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import importlib
import itertools
import time
import warnings
from functools import partial
from random import getrandbits
from typing import AsyncGenerator
from weakref import WeakSet

from libc.stdint cimport uint_fast64_t
from libc.stdlib cimport free, malloc

from .core cimport ActorRef, LocalActorRef
from .libcpp cimport mt19937_64


cdef mt19937_64 _rnd_gen
cdef bint _rnd_is_seed_set = False
NamedType = collections.namedtuple("NamedType", ["name", "type_"])
_type_dispatchers = WeakSet()

cdef class TypeDispatcher:
    def __init__(self):
        self._handlers = dict()
        self._lazy_handlers = dict()
        # store inherited handlers to facilitate unregistering
        self._inherit_handlers = dict()

        _type_dispatchers.add(self)

    cpdef void register(self, object type_, object handler):
        if isinstance(type_, str):
            self._lazy_handlers[type_] = handler
        elif type(type_) is not NamedType and isinstance(type_, tuple):
            for t in type_:
                self.register(t, handler)
        else:
            self._handlers[type_] = handler

    cpdef void unregister(self, object type_):
        if type(type_) is not NamedType and isinstance(type_, tuple):
            for t in type_:
                self.unregister(t)
        else:
            self._lazy_handlers.pop(type_, None)
            self._handlers.pop(type_, None)
            self._inherit_handlers.clear()

    cdef _reload_lazy_handlers(self):
        for k, v in self._lazy_handlers.items():
            mod_name, obj_name = k.rsplit('.', 1)
            with warnings.catch_warnings():
                # the lazy imported cudf will warn no device found,
                # when we set visible device to -1 for CPU processes,
                # ignore the warning to not distract users
                warnings.simplefilter("ignore")
                mod = importlib.import_module(mod_name, __name__)
            self.register(getattr(mod, obj_name), v)
        self._lazy_handlers = dict()

    cpdef get_handler(self, object type_):
        try:
            return self._handlers[type_]
        except KeyError:
            pass

        try:
            return self._inherit_handlers[type_]
        except KeyError:
            self._reload_lazy_handlers()
            if type(type_) is NamedType:
                named_type = partial(NamedType, type_.name)
                mro = itertools.chain(
                    *zip(map(named_type, type_.type_.__mro__),
                         type_.type_.__mro__)
                )
            else:
                mro = type_.__mro__
            for clz in mro:
                # only lookup self._handlers for mro clz
                handler = self._handlers.get(clz)
                if handler is not None:
                    self._inherit_handlers[type_] = handler
                    return handler
            raise KeyError(f'Cannot dispatch type {type_}')

    def __call__(self, object obj, *args, **kwargs):
        return self.get_handler(type(obj))(obj, *args, **kwargs)

    @staticmethod
    def reload_all_lazy_handlers():
        for dispatcher in _type_dispatchers:
            (<TypeDispatcher>dispatcher)._reload_lazy_handlers()


cpdef str to_str(s, encoding='utf-8'):
    if type(s) is str:
        return <str>s
    elif isinstance(s, bytes):
        return (<bytes>s).decode(encoding)
    elif isinstance(s, str):
        return str(s)
    elif s is None:
        return s
    else:
        raise TypeError(f"Could not convert from {s} to str.")


cpdef bytes to_binary(s, encoding='utf-8'):
    if type(s) is bytes:
        return <bytes>s
    elif isinstance(s, unicode):
        return (<unicode>s).encode(encoding)
    elif isinstance(s, bytes):
        return bytes(s)
    elif s is None:
        return None
    else:
        raise TypeError(f"Could not convert from {s} to bytes.")


cpdef void reset_id_random_seed() except *:
    cdef bytes seed_bytes
    global _rnd_is_seed_set

    seed_bytes = getrandbits(64).to_bytes(8, "little")
    _rnd_gen.seed((<uint_fast64_t *><char *>seed_bytes)[0])
    _rnd_is_seed_set = True


cpdef bytes new_random_id(int byte_len):
    cdef uint_fast64_t *res_ptr
    cdef uint_fast64_t res_data[4]
    cdef int i, qw_num = byte_len >> 3
    cdef bytes res

    if not _rnd_is_seed_set:
        reset_id_random_seed()

    if (qw_num << 3) < byte_len:
        qw_num += 1

    if qw_num <= 4:
        # use stack memory to accelerate
        res_ptr = res_data
    else:
        res_ptr = <uint_fast64_t *>malloc(qw_num << 3)

    try:
        for i in range(qw_num):
            res_ptr[i] = _rnd_gen()
        return <bytes>((<char *>&(res_ptr[0]))[:byte_len])
    finally:
        # free memory if allocated by malloc
        if res_ptr != res_data:
            free(res_ptr)


cpdef bytes new_actor_id():
    return new_random_id(32)


cdef set _is_async_generator_typecache = set()


cdef bint is_async_generator(obj):
    cdef type tp = type(obj)
    if tp in _is_async_generator_typecache:
        return True

    if isinstance(obj, AsyncGenerator):
        if len(_is_async_generator_typecache) < 100:
            _is_async_generator_typecache.add(tp)
        return True
    else:
        return False


def create_actor_ref(*args, **kwargs):
    """
    Create an actor reference.

    Returns
    -------
    ActorRef
    """

    cdef str address
    cdef object uid
    cdef ActorRef existing_ref

    address = to_str(kwargs.pop('address', None))
    uid = kwargs.pop('uid', None)
    proxy_addresses = kwargs.pop("proxy_addresses", None)

    if kwargs:
        raise ValueError('Only `address` or `uid` keywords are supported')

    # args: [address, uid], or [address, uid, proxy_addresses]
    if 2 <= len(args) <= 3 and isinstance(args[0], (str, bytes)):
        if address:
            raise ValueError('address has been specified')
        address = to_str(args[0])
        uid = args[1]
        if len(args) == 3:
            proxy_addresses = args[2]
    elif len(args) == 1:
        tp0 = type(args[0])
        if tp0 is ActorRef or tp0 is LocalActorRef:
            existing_ref = <ActorRef>(args[0])
            uid = existing_ref.uid
            address = to_str(address or existing_ref.address)
            proxy_addresses = existing_ref.proxy_addresses
        else:
            uid = args[0]

    if uid is None:
        raise ValueError('Actor uid should be provided')

    return ActorRef(address, uid, proxy_addresses=proxy_addresses)


cdef class Timer:
    def __enter__(self):
        self._start = time.time()
        return self

    def __exit__(self, *_):
        self.duration = time.time() - self._start



import os
import threading

_fury = threading.local()
_fury_not_installed = object()
_register_class_list = set()


def register_classes(*args):
    instance = get_fury()
    if instance is not None:
        _register_class_list.update(args)
        for c in _register_class_list:
            instance.register_class(c)


def get_fury():
    if os.environ.get("USE_FURY") in ("1", "true", "True"):
        instance = getattr(_fury, "instance", None)
        if instance is _fury_not_installed:  # pragma: no cover
            return None
        if instance is not None:
            return instance
        else:
            try:
                import pyfury

                _fury.instance = instance = pyfury.Fury(
                    language=pyfury.Language.PYTHON, require_class_registration=False
                )
                for c in _register_class_list:  # pragma: no cover
                    instance.register_class(c)
                print("pyfury is enabled.")
            except ImportError:  # pragma: no cover
                print("pyfury is not installed.")
                _fury.instance = _fury_not_installed
            return instance



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import re
import subprocess
import sys
import sysconfig
from distutils.command.build_ext import build_ext as _du_build_ext
from distutils.file_util import copy_file, move_file
from pathlib import Path

from sysconfig import get_config_vars

import numpy as np
from Cython.Build import cythonize
from packaging.version import Version
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext
from setuptools.command.install_lib import install_lib
from setuptools.extension import Library

try:
    import distutils.ccompiler

    if sys.platform != "win32":
        from numpy.distutils.ccompiler import CCompiler_compile

        distutils.ccompiler.CCompiler.compile = CCompiler_compile
except ImportError:
    pass

try:
    # Attempt to use Cython for building extensions, if available
    from Cython.Distutils.build_ext import build_ext as _build_ext

    # Additionally, assert that the compiler module will load
    # also. Ref #1229.
    __import__('Cython.Compiler.Main')
except ImportError:
    _build_ext = _du_build_ext

# From https://github.com/pandas-dev/pandas/pull/24274:
# For mac, ensure extensions are built for macos 10.9 when compiling on a
# 10.9 system or above, overriding distuitls behaviour which is to target
# the version that python was built for. This may be overridden by setting
# MACOSX_DEPLOYMENT_TARGET before calling setup.py
if sys.platform == "darwin":
    if "MACOSX_DEPLOYMENT_TARGET" not in os.environ:
        current_system = platform.mac_ver()[0]
        python_target = get_config_vars().get(
            "MACOSX_DEPLOYMENT_TARGET", current_system
        )
        target_macos_version = "10.9"

        parsed_python_target = Version(python_target)
        parsed_current_system = Version(current_system)
        parsed_macos_version = Version(target_macos_version)
        if parsed_python_target <= parsed_macos_version <= parsed_current_system:
            os.environ["MACOSX_DEPLOYMENT_TARGET"] = target_macos_version


repo_root = os.path.dirname(os.path.abspath(__file__))
os.chdir(repo_root)


cythonize_kw = dict(language_level=sys.version_info[0])
define_macros = []
if os.environ.get("CYTHON_TRACE"):
    define_macros.append(("CYTHON_TRACE_NOGIL", "1"))
    define_macros.append(("CYTHON_TRACE", "1"))
    cythonize_kw["compiler_directives"] = {"linetrace": True}

# Fixes Python 3.11 compatibility issue
#
# see also:
# - https://github.com/cython/cython/issues/4500
# - https://github.com/scoder/cython/commit/37f270155dbb5907435a1e83fdc90e403d776af5
if sys.version_info >= (3, 11):
    define_macros.append(("CYTHON_FAST_THREAD_STATE", "0"))

cy_extension_kw = {
    'define_macros': define_macros,
}

if "MSC" in sys.version:
    extra_compile_args = ["/std:c11", "/Ot", "/I" + os.path.join(repo_root, "misc")]
    cy_extension_kw["extra_compile_args"] = extra_compile_args
else:
    extra_compile_args = ["-O3"]
    if sys.platform != "darwin":
        # for macOS, we assume that C++ 11 is enabled by default
        extra_compile_args.append("-std=c++0x")
    cy_extension_kw["extra_compile_args"] = extra_compile_args


def _discover_pyx():
    exts = dict()
    for root, _, files in os.walk(os.path.join(repo_root, "xoscar"), followlinks=True):
        for fn in files:
            if not fn.endswith(".pyx"):
                continue
            full_fn = os.path.relpath(os.path.join(root, fn), repo_root)
            mod_name = full_fn.replace(".pyx", "").replace(os.path.sep, ".")
            exts[mod_name] = Extension(mod_name, [full_fn], include_dirs=[np.get_include()], **cy_extension_kw)
    return exts


extensions_dict = _discover_pyx()
cy_extensions = list(extensions_dict.values())

extensions = cythonize(cy_extensions, **cythonize_kw)


# Resolve path issue of versioneer
sys.path.append(repo_root)
versioneer = __import__("versioneer")


# build long description
def build_long_description():
    readme_path = os.path.join(os.path.dirname(os.path.abspath(repo_root)), "README.md")

    with open(readme_path, encoding="utf-8") as f:
        return f.read()


# Convert distutils Windows platform specifiers to CMake -A arguments
PLAT_TO_CMAKE = {
    "win32": "Win32",
    "win-amd64": "x64",
    "win-arm32": "ARM",
    "win-arm64": "ARM64",
}

TARGET_TO_PLAT = {
    'x86': 'win32',
    'x64': 'win-amd64',
    'arm': 'win-arm32',
    'arm64': 'win-arm64',
}


# Copied from https://github.com/pypa/setuptools/blob/main/setuptools/_distutils/util.py#L50
def get_host_platform():
    """
    Return a string that identifies the current platform. Use this
    function to distinguish platform-specific build directories and
    platform-specific built distributions.
    """

    # This function initially exposed platforms as defined in Python 3.9
    # even with older Python versions when distutils was split out.
    # Now it delegates to stdlib sysconfig, but maintains compatibility.
    return sysconfig.get_platform()


def get_platform():
    if os.name == 'nt':
        target = os.environ.get('VSCMD_ARG_TGT_ARCH')
        return TARGET_TO_PLAT.get(target) or get_host_platform()
    return get_host_platform()


plat_specifier = ".{}-{}".format(get_platform(), sys.implementation.cache_tag)


def get_build_lib():
    return os.path.join("build", "lib" + plat_specifier)


def get_build_temp():
    return os.path.join("build", 'temp' + plat_specifier)


# A CMakeExtension needs a sourcedir instead of a file list.
# The name must be the _single_ output extension from the CMake build.
# If you need multiple extensions, see scikit-build.
class XoscarCmakeExtension(Extension):
    def __init__(self, name: str, sourcedir: str = "") -> None:
        super().__init__(name, sources=[])
        self.sourcedir = os.fspath(Path(sourcedir).resolve())


class CMakeBuild(build_ext):
    def finalize_options(self):
        """
        For python 3.12, the build_temp and build_lib dirs are temp dirs which are depended on your OS,
        which leads to that cannot find the copy directory during C++ compiled process.
        However, for Python < 3.12, these two dirs can be automatically located in the `build` directory of the project directory.
        Therefore, in order to be compatible with all Python versions,
        directly using fixed dirs here by coping source codes from `setuptools`.
        """
        self.build_temp = get_build_temp()
        self.build_lib = get_build_lib()
        super().finalize_options()

    def copy_extensions_to_source(self):
        build_py = self.get_finalized_command('build_py')
        for ext in self.extensions:
            if not isinstance(ext, XoscarCmakeExtension):
                fullname = self.get_ext_fullname(ext.name)
                filename = self.get_ext_filename(fullname)
                modpath = fullname.split('.')
                package = '.'.join(modpath[:-1])
                package_dir = build_py.get_package_dir(package)
                dest_filename = os.path.join(package_dir,
                                             os.path.basename(filename))
                src_filename = os.path.join(self.build_lib, filename)

                # Always copy, even if source is older than destination, to ensure
                # that the right extensions for the current Python/platform are
                # used.
                copy_file(
                    src_filename, dest_filename, verbose=self.verbose,
                    dry_run=self.dry_run
                )
                if ext._needs_stub:
                    self.write_stub(package_dir or os.curdir, ext, True)
            else:
                fullname = self.get_ext_fullname(ext.name)
                collective_dir = os.path.join("xoscar" , "collective")
                filename = self.get_ext_filename(fullname)
                src_dir = os.path.join(self.build_lib , collective_dir)
                src_filename = os.path.join(src_dir , filename)
                dest_filename = os.path.join(collective_dir,
                                                os.path.basename(filename))
                copy_file(
                    src_filename, dest_filename, verbose=self.verbose,
                    dry_run=self.dry_run
                )


    def build_extension(self, ext):
        # TODO: support windows compilation
        is_windows = sys.platform.startswith('win')
        bit_number = platform.architecture()[0]
        if isinstance(ext, XoscarCmakeExtension) and not (is_windows and bit_number=="32bit"):
            self.build_Cmake(ext)
        elif isinstance(ext, XoscarCmakeExtension) and (is_windows and bit_number=="32bit"):
            pass
        else:
            ext._convert_pyx_sources_to_lang()
            _compiler = self.compiler
            try:
                if isinstance(ext, Library):
                    self.compiler = self.shlib_compiler
                _build_ext.build_extension(self, ext)
                if ext._needs_stub:
                    build_lib = self.get_finalized_command('build_py').build_lib
                    self.write_stub(build_lib, ext)
            finally:
                self.compiler = _compiler

    def build_Cmake(self, ext: XoscarCmakeExtension) -> None:
        # Must be in this form due to bug in .resolve() only fixed in Python 3.10+
        ext_fullpath = Path.cwd() / self.get_ext_fullpath(ext.name)
        extdir = ext_fullpath.parent.resolve()
        source_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
        build_temp = Path(self.build_temp) / ext.name
        if not build_temp.exists():
            build_temp.mkdir(parents=True)

        # Using this requires trailing slash for auto-detection & inclusion of
        # auxiliary "native" libs

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        cfg = "Debug" if debug else "Release"

        # CMake lets you override the generator - we need to check this.
        # Can be set with Conda-Build, for example.
        cmake_generator = os.environ.get("CMAKE_GENERATOR", "")

        # Set Python_EXECUTABLE instead if you use PYBIND11_FINDPYTHON
        # EXAMPLE_VERSION_INFO shows you how to pass a value into the C++ code
        # from Python.
        cmake_args = [
            f"-DBUILD_TMP_DIR={build_temp}",
            f"-DPYTHON_PATH={sys.executable}",
            f"-DCMAKE_BUILD_TYPE={cfg}",  # not used on MSVC, but no harm
            "-DCMAKE_POLICY_VERSION_MINIMUM=3.10",
        ]
        build_args = []
        # Adding CMake arguments set as environment variable
        # (needed e.g. to build for ARM OSx on conda-forge)
        if "CMAKE_ARGS" in os.environ:
            cmake_args += [item for item in os.environ["CMAKE_ARGS"].split(" ") if item]

        if self.compiler.compiler_type != "msvc":
            # Using Ninja-build since it a) is available as a wheel and b)
            # multithreads automatically. MSVC would require all variables be
            # exported for Ninja to pick it up, which is a little tricky to do.
            # Users can override the generator with CMAKE_GENERATOR in CMake
            # 3.15+.
            if not cmake_generator or cmake_generator == "Ninja":
                try:
                    import ninja

                    ninja_executable_path = Path(ninja.BIN_DIR) / "ninja"
                    cmake_args += [
                        "-GNinja",
                        f"-DCMAKE_MAKE_PROGRAM:FILEPATH={ninja_executable_path}",
                    ]
                except ImportError:
                    pass

        else:
            # Single config generators are handled "normally"
            single_config = any(x in cmake_generator for x in {"NMake", "Ninja"})

            # CMake allows an arch-in-generator style for backward compatibility
            contains_arch = any(x in cmake_generator for x in {"ARM", "Win64"})

            # Specify the arch if using MSVC generator, but only if it doesn't
            # contain a backward-compatibility arch spec already in the
            # generator name.
            if not single_config and not contains_arch:
                cmake_args += ["-A", PLAT_TO_CMAKE[self.plat_name]]

            # Multi-config generators have a different way to specify configs
            if not single_config:
                cmake_args += [
                    f"-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{cfg.upper()}={extdir}"
                ]
                build_args += ["--config", cfg]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        # Set CMAKE_BUILD_PARALLEL_LEVEL to control the parallel build level
        # across all generators.
        if "CMAKE_BUILD_PARALLEL_LEVEL" not in os.environ:
            # self.parallel is a Python 3 only way to set parallel jobs by hand
            # using -j in the build_ext call, not supported by pip or PyPA-build.
            if hasattr(self, "parallel") and self.parallel:
                # CMake 3.12+ only.
                build_args += [f"-j{self.parallel}"]



        subprocess.run(
            ["cmake", source_dir, *cmake_args], cwd=build_temp, check=True
        )
        subprocess.run(
            ["cmake", "--build", ".", *build_args], cwd=build_temp, check=True
        )
        if sys.platform.startswith('win'):
            for file in os.listdir(self.build_lib):
                if file.startswith("xoscar_pygloo"):
                    src_filename = os.path.join(self.build_lib,
                                            os.path.basename(file))
                    dest_dir = os.path.join(self.build_lib,
                                                    "xoscar\\collective")
                    if not os.path.exists(dest_dir):
                        os.mkdir(dest_dir)
                    dest_filename = os.path.join(dest_dir,
                                            os.path.basename(file))
                    move_file(
                        src_filename, dest_filename, verbose=self.verbose,
                        dry_run=self.dry_run
                    )
                    libuv_filename = "xoscar\\collective\\uv.dll"
                    libuv_dest_filename = os.path.join(dest_dir, "uv.dll")
                    copy_file(
                        libuv_filename, libuv_dest_filename, verbose=self.verbose,
                        dry_run=self.dry_run
                    )


setup_options = dict(
    version=versioneer.get_version(),
    ext_modules=extensions + [XoscarCmakeExtension("xoscar_pygloo")],
    cmdclass={"build_ext": CMakeBuild},
    long_description=build_long_description(),
    long_description_content_type="text/markdown",
)
setup(**setup_options)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


cdef class ActorRef:
    cdef object __weakref__
    cdef public str address
    cdef public object uid
    cdef public list proxy_addresses
    cdef dict _methods


cdef class LocalActorRef(ActorRef):
    cdef object _actor_weakref
    cdef _weakref_local_actor(self)


cdef class BufferRef:
    cdef public str address
    cdef public bytes uid


cdef class FileObjectRef:
    cdef public str address
    cdef public bytes uid


cdef class _BaseActor:
    cdef object __weakref__
    cdef str _address
    cdef object _lock
    cdef object _uid

    cpdef ActorRef ref(self)


cdef class ActorEnvironment:
    cdef public dict actor_locks
    cdef public object address
